/*
 * unity-webapps-script-repo.c
 * Copyright (C) Canonical LTD 2012
 * 
 * Author: Alexandre Abreu <alexandre.abreu@canonical.com>
 * 
 unity-webapps is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * unity-webapps is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.";
 */

/**
 * SECTION:unity-webapps-script-repo
 * @short_description: Unity Webapps user script repository handling & exposure
 *
 * Unity Webapps user script repository handling & exposure
 * * readonly view of the script repo
 */

#include <glib.h>
#include <gio/gio.h>
#include <string.h>

#include "unity-webapps-script-repo.h"

static void free_includes_list (UnityWebappsIntegrationScriptDescription * pDescription);


static gchar * get_user_defined_repo_path (void)
{
  const gchar * UNITY_WEBAPPS_SCRIPT_REPO_ENVVARNAME = "UNITY_WEBAPPS_SCRIPT_REPO_PATH";

  gchar * repo_path = g_strdup (g_getenv (UNITY_WEBAPPS_SCRIPT_REPO_ENVVARNAME));

  return repo_path;
}

static gchar * expand_path (const gchar * const path)
{
  const gchar * HOME_SHORTCUT = "~";
  if (g_str_has_prefix (path, HOME_SHORTCUT))
    {
      return g_build_path (G_DIR_SEPARATOR_S, g_get_home_dir (), path + strlen(HOME_SHORTCUT), NULL);
    }
  return g_strdup (path);
}
  
static GSList * get_list_of_search_repo_path (void)
{
  // ordered in search order
  gchar * ordered_repo_path_search_list [] = {
    "~/.local/share"
    , "/usr/local/share"
    , "/usr/share"
  };
  
  GSList * path_list = NULL;
  
  gchar * env_repo_path = get_user_defined_repo_path ();

  // The presence of the environment variable takes over the rest.
  // Mostly used for testing.
  if (NULL != env_repo_path && 0 != strlen (env_repo_path))
    {
      path_list = g_slist_prepend (path_list, env_repo_path);
      return path_list;
    }

  const gchar * UNITY_WEBAPPS_INSTALLED_REPO_FOLDER_NAME = "unity-webapps/userscripts";

  gsize i = 0;
  for (; i < G_N_ELEMENTS(ordered_repo_path_search_list); ++i)
    {
      gchar * expanded_base_path = expand_path (ordered_repo_path_search_list[i]);
      
      gchar * path = g_build_path (G_DIR_SEPARATOR_S
				   , expanded_base_path
				   , UNITY_WEBAPPS_INSTALLED_REPO_FOLDER_NAME
				   , NULL);
      
      g_free (expanded_base_path);
      
      path_list = g_slist_append (path_list, path);
      
#if 0 // disabled for now since might not be handled here
      if (is_proper_folder (path))
	{
	}
      else
	{
	  g_message ("'%s' does not exist - not considered as a userscript source path", path);
	  g_free (path);
	}
#endif
    }
  
  return path_list;
}

static void do_free_path (gpointer data, gpointer user_data)
{
  g_free (data);
}

static void free_repo_path_list (GSList * list)
{
  g_return_if_fail (NULL != list);
  g_slist_foreach (list, do_free_path, NULL);
  g_slist_free (list);
}

static gchar * normalize_repository_path (gchar * const path)
{
  g_return_val_if_fail (NULL != path, NULL);
  g_return_val_if_fail (0 != strlen(path), path);

  gchar * normalized_path = g_strdup (path);

  // normalize
  g_strstrip (normalized_path);
  if (g_str_has_suffix (normalized_path, G_DIR_SEPARATOR_S))
    {
      normalized_path [strlen (normalized_path)] = '\0';
    }

  return normalized_path;
}

static const gchar * USER_SCRIPT_DEFINITION_START = "==UserScript==";
static const gchar * USER_SCRIPT_DEFINITION_END = "==/UserScript==";
static const gchar * USER_SCRIPT_DEFINITION_LINE_START = "//";
static const gchar * USER_SCRIPT_NAME_DIRECTIVE = "@name";
static const gchar * USER_SCRIPT_NAMESPACE_DIRECTIVE = "@namespace";
static const gchar * USER_SCRIPT_DESCRIPTION_DIRECTIVE = "@description";
static const gchar * USER_SCRIPT_INCLUDE_DIRECTIVE = "@include";
static const gchar * USER_SCRIPT_VERSION_DIRECTIVE = "@version";
static const gchar * USER_SCRIPT_AUTHOR_DIRECTIVE = "@author";
static const gchar * USER_SCRIPT_REQUIRE_DIRECTIVE = "@require";

#define SCRIPT_PARSE_DEFINE_SCRIPT_DESCRIPTION_PREDICATE(name, description_string) \
  static gboolean script_parse_is_##name (const gchar * const line)	\
  {									\
    gchar * dline = g_strdup (line);					\
    g_strstrip (dline);							\
									\
    if ( ! g_str_has_prefix (dline, USER_SCRIPT_DEFINITION_LINE_START)) \
      {									\
	return FALSE;							\
      }									\
									\
    gchar * rest = dline + strlen(USER_SCRIPT_DEFINITION_LINE_START);	\
    g_strstrip (rest);							\
									\
    gboolean result = g_str_has_prefix (rest, description_string)	\
      && (	g_ascii_isspace (rest[strlen(description_string)])	\
		|| 	'\0' == rest[strlen(description_string)] );	\
									\
    g_free (dline);							\
									\
    return result;							\
  }

// defines script_parse_is_definition_start
SCRIPT_PARSE_DEFINE_SCRIPT_DESCRIPTION_PREDICATE (definition_start, USER_SCRIPT_DEFINITION_START);

// defines script_parse_is_definition_end
SCRIPT_PARSE_DEFINE_SCRIPT_DESCRIPTION_PREDICATE (definition_end, USER_SCRIPT_DEFINITION_END);

// defines script_parse_is_name_directive
SCRIPT_PARSE_DEFINE_SCRIPT_DESCRIPTION_PREDICATE (name_directive, USER_SCRIPT_NAME_DIRECTIVE);

// defines script_parse_is_namespace_directive
SCRIPT_PARSE_DEFINE_SCRIPT_DESCRIPTION_PREDICATE (namespace_directive, USER_SCRIPT_NAMESPACE_DIRECTIVE);

// defines script_parse_is_description_directive
SCRIPT_PARSE_DEFINE_SCRIPT_DESCRIPTION_PREDICATE (description_directive, USER_SCRIPT_DESCRIPTION_DIRECTIVE);

// defines script_parse_is_include_directive
SCRIPT_PARSE_DEFINE_SCRIPT_DESCRIPTION_PREDICATE (include_directive, USER_SCRIPT_INCLUDE_DIRECTIVE);

// defines script_parse_is_version_directive
SCRIPT_PARSE_DEFINE_SCRIPT_DESCRIPTION_PREDICATE (version_directive, USER_SCRIPT_VERSION_DIRECTIVE);

// defines script_parse_is_author_directive
SCRIPT_PARSE_DEFINE_SCRIPT_DESCRIPTION_PREDICATE (author_directive, USER_SCRIPT_AUTHOR_DIRECTIVE);

// defines script_parse_is_require_directive
SCRIPT_PARSE_DEFINE_SCRIPT_DESCRIPTION_PREDICATE (require_directive, USER_SCRIPT_REQUIRE_DIRECTIVE);


static const gchar * skip_spaces (const gchar * s)
{
  if (NULL != s)
    {
      // TODO SERIOUS TODO ... should be UTF8 or something? (user defined string)
      while ('\0' != *s && g_ascii_isspace (*s)) ++s;
    }
  
  return s;
}


/**
 *
 * @return the directive data
 * 		   NULL if any error
 */
static const gchar * script_parse_extract_directive_data (const gchar * const line
							  , const gchar * const directive_name)
{
  const gchar * curline = line;
  curline = skip_spaces(curline);

  if ( ! g_str_has_prefix (curline, USER_SCRIPT_DEFINITION_LINE_START))
    {
      g_critical ("Invalid user definition line %s", line);
      return NULL;
    }
  curline += strlen(USER_SCRIPT_DEFINITION_LINE_START);

  curline = skip_spaces(curline);
  if ( ! g_str_has_prefix (curline, directive_name))
    {
      g_critical ("Invalid user definition line (expected directive %s) %s", directive_name, line);
      return NULL;
    }
  curline += strlen(directive_name);

  curline = skip_spaces(curline);

  return curline;
}

/**
 *
 * @param complete file path of the required script
 * @param name of the required script (a bit redundant)
 * @param do_fail_on_content_load_fail determined if the function returns NULL on error (script content load op)
 * 
 * @return integration script structure filled w/ content of the js file and name
 */
static UnityWebappsIntegrationScriptContent * load_require_script_by_name (const gchar * const filename
									   , const gchar * const name
									   , gboolean do_fail_on_content_load_fail)
{
  UnityWebappsIntegrationScriptContent * pContent = NULL;

  gchar * content = NULL;
  gsize length = 0;

  if ( ! g_file_get_contents (filename, &content, &length, NULL) && do_fail_on_content_load_fail)
    {
      g_critical ("Could not load require script: %s", filename);
      return NULL;
    }	

  pContent = g_malloc0 (sizeof(*pContent));
  g_return_val_if_fail (NULL != pContent, NULL);

  pContent->name = g_strdup (name);
  pContent->content = content;

  return pContent;
}


// TODO leaks script_parse_extract_directive_data results!!!!!!!!!!!!!!!!
// return true if continue
static gboolean script_parse_directive_line (UnityWebappsIntegrationScriptDescription * pDescription
					     , const gchar * const line
					     , const gchar * repo_path)
{
  if (script_parse_is_definition_end (line))
    {
      return FALSE;
    }

  if (script_parse_is_name_directive (line))
    {
      if (NULL != pDescription->name && 0 != strlen (pDescription->name))
	{
	  g_warning ("Duplicate name definition found");
	  g_free (pDescription->name);
	}
      pDescription->name = g_strdup (script_parse_extract_directive_data (line, USER_SCRIPT_NAME_DIRECTIVE));
    }

  if (script_parse_is_namespace_directive (line))
    {
      if (NULL != pDescription->namespace && 0 != strlen (pDescription->namespace))
	{
	  g_warning ("Duplicate namespace definition found");
	  g_free (pDescription->namespace);
	}
      pDescription->namespace = g_strdup (script_parse_extract_directive_data (line, USER_SCRIPT_NAMESPACE_DIRECTIVE));
    }
	
  if (script_parse_is_description_directive (line))
    {
      if (NULL != pDescription->description && 0 != strlen (pDescription->description))
	{
	  g_warning ("Duplicate description definition found");
	  g_free (pDescription->description);
	}
      pDescription->description = g_strdup (script_parse_extract_directive_data (line, USER_SCRIPT_DESCRIPTION_DIRECTIVE));
    }
	
  if (script_parse_is_include_directive (line))
    {
      pDescription->includes = g_slist_append (pDescription->includes
					       , g_strdup (script_parse_extract_directive_data (line, USER_SCRIPT_INCLUDE_DIRECTIVE))		
					       );
    }
	
  if (script_parse_is_version_directive (line))
    {
      if (NULL != pDescription->version && 0 != strlen (pDescription->version))
	{
	  g_warning ("Duplicate version definition found");
	  g_free (pDescription->version);
	}
      pDescription->version = g_strdup (script_parse_extract_directive_data (line, USER_SCRIPT_VERSION_DIRECTIVE));
    }

  if (script_parse_is_author_directive (line))
    {
      if (NULL != pDescription->author && 0 != strlen (pDescription->author))
	{
	  g_warning ("Duplicate author definition found");
	  g_free (pDescription->author);
	}
      pDescription->author = g_strdup (script_parse_extract_directive_data (line, USER_SCRIPT_AUTHOR_DIRECTIVE));
    }
	
  if (script_parse_is_require_directive (line))
    {
      const gchar * require_script_name = script_parse_extract_directive_data (line, USER_SCRIPT_REQUIRE_DIRECTIVE);
      gchar * full_script_path = g_build_path (G_DIR_SEPARATOR_S
					       , repo_path
					       , require_script_name
					       , NULL);

      // we allow partial (e.g. when we failed loading the required script content)
      //	failure, a validation step will afterward fix this and properly reject this script
      UnityWebappsIntegrationScriptContent *
	pContent = load_require_script_by_name (full_script_path
						, require_script_name
						, FALSE // do not return NULL if content load fails
						);
      
      if (NULL != pContent)
	{
	  pDescription->requires = g_slist_append (pDescription->requires
						   , pContent
						   );
	}

      g_free (full_script_path);
    }

  return TRUE;
}

static gboolean script_parse_definition_start (UnityWebappsIntegrationScriptDescription * pDescription
					       , gchar ** lines
					       , guint * pLineIdx
					       , guint line_count
					       , const gchar * repo_path)
{
  g_return_val_if_fail (NULL != lines, FALSE);
  g_return_val_if_fail (NULL != pDescription, FALSE);
  g_return_val_if_fail (NULL != pLineIdx, FALSE);


  if ( ! script_parse_is_definition_start (lines[*pLineIdx]))
    {
      return FALSE;
    }

#define NEXT_LINE() (*pLineIdx)++
#define GET_LINE() (((*pLineIdx) >= line_count) ? NULL : lines[*pLineIdx])
#define SKIP_EMPTY_LINES() \
  { \
    while ( 	((*pLineIdx) < line_count) \
		&& 	0 == strlen (g_strstrip (lines[*pLineIdx])) ) \
      { \
	(*pLineIdx)++; \
      } \
  }
  
  NEXT_LINE();
  SKIP_EMPTY_LINES();
  
  while (((*pLineIdx) < line_count) && script_parse_directive_line (pDescription, GET_LINE(), repo_path))
    {
      NEXT_LINE();
      SKIP_EMPTY_LINES();
    }
  
#undef NEXT_LINE
#undef GET_LINE
#undef SKIP_EMPTY_LINES
  
  return TRUE;
}


static UnityWebappsIntegrationScriptDescription *
script_parse_content (const gchar * const content, const gchar * repo_path)
{
  UnityWebappsIntegrationScriptDescription * pDescription = NULL;

  gchar ** lines = g_strsplit (content, "\n", 0);
  g_return_val_if_fail (NULL != lines, pDescription);

  guint line_count = 0;
  while (lines[line_count] != NULL) line_count++;

  g_return_val_if_fail (0 != line_count, pDescription);
	
  guint lineIdx = 0;

  // 
#define SKIP_EMPTY_LINES() \
  { \
    while ( 	(lineIdx < line_count) \
		&& 	0 == strlen (g_strstrip (lines[lineIdx])) ) \
      { \
	lineIdx++; \
      } \
  }

  SKIP_EMPTY_LINES ();
  if (lineIdx >= line_count)
    {
      // reached the end
      g_strfreev (lines);
      return pDescription;
    }

  pDescription = g_malloc0 (sizeof (UnityWebappsIntegrationScriptDescription));
  pDescription->content = g_strdup(content);

  // some sort of fake recursive descent ... 
  script_parse_definition_start (pDescription, lines, &lineIdx, line_count, repo_path);

  if (lineIdx >= line_count)
    {
      g_warning ("Ooops, went over the board w/ user scrip parsing");
    }

  g_strfreev (lines);

  return pDescription;
}



static const gchar * const SCRIPT_FILENAME_EXTENSION = ".user.js";
static GSList * list_script_files_in_dir (const gchar * path, const gchar * script_file_extension)
{
  g_return_val_if_fail (NULL != path, NULL);
  g_return_val_if_fail (NULL != script_file_extension, NULL);
  g_return_val_if_fail (0 != strlen (script_file_extension), NULL);
  g_return_val_if_fail (0 != strlen (path), NULL);
  
  g_return_val_if_fail (g_file_test (path, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR), NULL);
  
  // TODO mmh maybe trim script_file_extension

  GDir * dir = g_dir_open (path, 0, NULL);
  if (NULL == dir)
    {
      return NULL;
    }
  
  GSList * list = NULL;

  const gchar * filename = g_dir_read_name (dir);
  while (NULL != filename)
    {
      if (g_str_has_suffix (filename, script_file_extension))
	{
	  list = g_slist_append (list, g_strdup (filename));
	  g_message ("File %s is being considered as a valid integration script", filename);
	}
      else
	{
	  g_message ("File %s is NOT being considered as a valid integration script", filename);
	}
      
      filename = g_dir_read_name (dir);
    }
  g_dir_close (dir);
  dir = NULL;

  return list;
}

static void free_include_list (gpointer data, gpointer user_data)
{
  g_free (data);
}

static void free_includes_list (UnityWebappsIntegrationScriptDescription * pDescription)
{
  g_return_if_fail (NULL != pDescription);

  if (NULL != pDescription->includes)
    {
      g_slist_foreach (pDescription->includes, free_include_list, NULL);
      g_slist_free (pDescription->includes);
    }
}

static void free_require_list (gpointer data, gpointer user_data)
{
  UnityWebappsIntegrationScriptContent * pContent = (UnityWebappsIntegrationScriptContent *) data;
  if (NULL != pContent)
    {
      g_free(pContent->name);
      g_free(pContent->content);
    }
  g_free (pContent);
}

static void free_integration_script_description (UnityWebappsIntegrationScriptDescription * pDescription)
{
  g_return_if_fail (NULL != pDescription);

  g_free (pDescription->name);
  g_free (pDescription->namespace);
  g_free (pDescription->description);
  g_free (pDescription->version);
  g_free (pDescription->author);
  g_free (pDescription->content);

  free_includes_list (pDescription);

  if (NULL != pDescription->requires)
    {
      g_slist_foreach (pDescription->requires, free_require_list, NULL);
      g_slist_free (pDescription->requires);
    }
}

static void check_includes_list (gpointer data, gpointer user_data)
{
  g_return_if_fail(NULL != user_data);
  
  gboolean * pValidIncludes = (gboolean *) user_data;
  if (FALSE == *pValidIncludes)
    {
      // we are already false, just skip
      return;
    }
  
  if (NULL == data)
    {
      *pValidIncludes = FALSE;
    }
}

static void check_requires_list (gpointer data, gpointer user_data)
{
  g_return_if_fail(NULL != user_data);
  
  gboolean * pValidRequires = (gboolean *) user_data;
  if (FALSE == *pValidRequires)
    {
      // we are already false, just skip
      return;
    }
  
  UnityWebappsIntegrationScriptContent *
    pContent = (UnityWebappsIntegrationScriptContent *) data;
  if (NULL == pContent)
    {
      *pValidRequires = FALSE;
    }
  else
    {
      *pValidRequires = (NULL != pContent->name) && (NULL != pContent->content);
    }
}

static gboolean is_valid_integration_script (UnityWebappsIntegrationScriptDescription * pDescription, gchar ** error)
{
  if (NULL == pDescription->name
      || 0 == strlen (pDescription->name))
    {
      if (NULL != error)
	{
	  *error = g_strdup ("Invalid script name");
	}
      return FALSE;
    }

  if (NULL == pDescription->content
      || 0 == strlen (pDescription->content))
    {
      if (NULL != error)
	{
	  *error = g_strdup ("Invalid script content");
	}
      return FALSE;
    }

  if (NULL == pDescription->version
      || 0 == strlen (pDescription->version)
      )
    {
      if (NULL != error)
	{
	  *error = g_strdup ("Invalid script version");
	}
      return FALSE;
    }

  // check includes
  if (NULL == pDescription->includes)
    {
      if (NULL != error)
	{
	  *error = g_strdup ("No includes found");
	}
      return FALSE;
    }

  if (0 == g_slist_length (pDescription->includes))
    {
      if (NULL != error)
	{
	  *error = g_strdup ("Empty includes list");
	}
      return FALSE;
    }

  gboolean are_valid_includes = TRUE;
  g_slist_foreach (pDescription->includes, check_includes_list, &are_valid_includes);

  if (FALSE == are_valid_includes)
    {
      if (NULL != error)
	{
	  *error = g_strdup ("Empty includes");
	}
      return are_valid_includes;
    }
	
  // check requires if any
  gboolean are_valid_requires = TRUE;

  if (	NULL != pDescription->requires
	&& 	0 != g_slist_length (pDescription->requires)	)
    {
      g_slist_foreach (pDescription->requires, check_requires_list, &are_valid_requires);
    }

  if (FALSE == are_valid_requires && NULL != error)
    {
      *error = g_strdup ("Invalid requires");
    }

  return are_valid_requires;
}


static gboolean
unity_webapps_script_repo_get_descriptions_for_path (UnityWebappsIntegrationScriptDescription ** ppDescriptions
						     , gsize descriptions_count
						     , gsize * pCurIdx
						     , gchar * script_path)
{
  g_return_val_if_fail (NULL != pCurIdx, FALSE);
  g_return_val_if_fail (NULL != script_path, FALSE);
  g_return_val_if_fail (NULL != ppDescriptions, FALSE);
  
  if (0 == g_file_test (script_path, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))
    {
      return TRUE;
    }
  
  // init
  gchar * normalized_script_path = normalize_repository_path (script_path);
  g_return_val_if_fail (NULL != normalized_script_path, FALSE);
  g_return_val_if_fail (0 != strlen (normalized_script_path), FALSE);

  GSList *
    script_list = list_script_files_in_dir (normalized_script_path, SCRIPT_FILENAME_EXTENSION);
  if (NULL == script_list)
    {
      g_free (normalized_script_path);
      return TRUE;
    }
  
  const guint USER_SCRIPT_COUNT = g_slist_length (script_list);
  g_return_val_if_fail (0 != USER_SCRIPT_COUNT, TRUE);
  
  // parse each user script
  guint idx = 0;
  GSList * cur_script = script_list;
  for (; NULL != cur_script; cur_script = g_slist_next(cur_script))
    {
      gpointer data = cur_script->data;
      
      if (idx >= USER_SCRIPT_COUNT)
	{
	  // OOPS
	  g_message ("Out of range while enumerating user scripts in folder '%s'", normalized_script_path);
	  continue;
	}
      
      if ((*pCurIdx + idx) >= descriptions_count)
	{
	  // OOPS
	  g_message ("Global out of range while enumerating user"
		     " scripts in folder '%s' ignoring '%s'"
		     , normalized_script_path
		     , (gchar*) data);
	  continue;
	}
      
      gchar * scriptname = (gchar *) data;
      gchar * fullscriptpath = g_strconcat (normalized_script_path, G_DIR_SEPARATOR_S, scriptname, NULL);
      
      gchar * content = NULL;
      gsize length = 0;
      
      g_message ("Parsing integration script %s", fullscriptpath);
      
      if ( ! g_file_get_contents (fullscriptpath, &content, &length, NULL))
	{
	  g_critical ("Could not load user script: %s", fullscriptpath);
	  g_free (fullscriptpath);
	  continue;
	}
      
      UnityWebappsIntegrationScriptDescription *
	pDescription = script_parse_content (content, script_path);
      
      gchar * error = NULL;
      if (is_valid_integration_script (pDescription, &error))
	{
	  g_message ("Adding userscript '%s'", scriptname);
	  
	  ppDescriptions[idx] = pDescription;
	  ++idx;
	}
      else
	{
	  g_message ("Ignoring userscript '%s'", scriptname);
	  
	  free_integration_script_description (pDescription);
	  g_free (error);
	}
      
      g_free (content);
      g_free (fullscriptpath);
      
    } // for (; NULL != cur_script

  if (idx != USER_SCRIPT_COUNT)
    {
      g_warning ("Some user scripts have been ignored (count %d) in folder '%s'"
		 , (USER_SCRIPT_COUNT - idx)
		 , normalized_script_path);
    }
	
  *pCurIdx += idx;
	
  // wrap up & leave
  cur_script = script_list;
  for (; NULL != cur_script; cur_script = g_slist_next(cur_script))
    {
      g_free (cur_script->data);
    }

  g_slist_free (script_list);
  g_free (normalized_script_path);

  return TRUE;
}



//////////////////////////////////////////////////////////////////////////////
//
//
//		public part
//
//
//////////////////////////////////////////////////////////////////////////////


#include <errno.h>


/**
 * 
 * @return EAGAIN if size is not big enough
 * 		   EINVAL any error
 */
gint unity_webapps_script_repo_get_path (gchar * path, guint size)
{
  gchar * script_path = NULL;

  GSList * list = get_list_of_search_repo_path ();

  GSList * cur_path = list;
  for (; NULL != cur_path; cur_path = g_slist_next(cur_path))
    {
      gpointer data = cur_path->data;
      
      if (NULL == data)
	{
	  continue;
	}
      
      if (NULL == script_path)
	{
	  script_path = g_strdup (data);
	}
      else
	{
	  gchar * old_script_path = script_path;
	  script_path = g_strjoin (";", (gchar *) data, old_script_path, NULL);
	  g_free (old_script_path);
	}
    }
  
  free_repo_path_list (list);

  g_return_val_if_fail (NULL != script_path, EINVAL);

  if (NULL != script_path && (strlen(script_path) > (size-1)))
    {
      g_free (script_path);
      return EAGAIN;
    }

  if (NULL != script_path)
    {
      g_strlcpy (path, script_path, size);
    }

  g_free (script_path);

  return 0;
}

guint get_script_count_in_path (const gchar * path)
{
  g_return_val_if_fail (NULL != path, 0);
  g_return_val_if_fail (0 != strlen (path), 0);
  
  if (0 == g_file_test (path, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))
    {
      g_message ("The unity webapps integration script search path %s does not exist", path);
      return 0;
    }
  
  GSList
    * list = list_script_files_in_dir (path, SCRIPT_FILENAME_EXTENSION);
  if (NULL == list)
    {
      return 0;
    }

  guint count = g_slist_length (list);
  g_slist_free (list);

  return count;
}

guint unity_webapps_script_repo_get_script_count (void)
{
  GSList * list = get_list_of_search_repo_path ();

  guint count = 0;

  GSList * cur_path = list;
  for (; NULL != cur_path; cur_path = g_slist_next(cur_path))
    {
      gpointer data = cur_path->data;
      count += get_script_count_in_path ((gchar *) data);
    }

  free_repo_path_list (list);

  return count;
}


/**
 * Extract the description for all integration scripts found in the repos
 * 
 * @return list of script description structures, is a NULL terminated array of pointers
 * @note the returned list *must* be freed w/ unity_webapps_script_repo_free_all
 */
UnityWebappsIntegrationScriptDescription ** unity_webapps_script_repo_get_descriptions (void)
{
  UnityWebappsIntegrationScriptDescription ** ppDescriptions = NULL;

  // 1. retrieve proper number of files
  const guint USER_SCRIPT_COUNT = unity_webapps_script_repo_get_script_count ();
  g_return_val_if_fail (0 != USER_SCRIPT_COUNT, ppDescriptions);

  ppDescriptions = g_malloc0 ((USER_SCRIPT_COUNT + 1) * sizeof (UnityWebappsIntegrationScriptDescription *));


  // 2. extract scripts for all of them
  gsize idx = 0;
	
  GSList * search_repo_paths = get_list_of_search_repo_path ();

  GSList * cur_path = search_repo_paths;
  for (; NULL != cur_path; cur_path = g_slist_next(cur_path))
    {
      gpointer data = cur_path->data;
      if (idx < USER_SCRIPT_COUNT)
	{
	  unity_webapps_script_repo_get_descriptions_for_path (ppDescriptions
							       , USER_SCRIPT_COUNT
							       , &idx
							       , (gchar *) data);
	}
      else
	{
	  g_message ("Integration script count mismatch: some scripts are being ignored");
	}
    }
  free_repo_path_list (search_repo_paths);

  return ppDescriptions;
}


/**
 * Frees some data allocated by a call to unity_webapps_script_repo_get_descriptions
 * 
 */
void unity_webapps_script_repo_free_all (UnityWebappsIntegrationScriptDescription ** ppDescriptions)
{
  g_return_if_fail (ppDescriptions != NULL);

  guint idx = 0;
  for (; ppDescriptions[idx] != NULL; ++idx)
    {
      free_integration_script_description (ppDescriptions[idx]);
    }

  g_free (ppDescriptions);
}





